//----------------------------------------------------------------------------
//
//  Simple Circuits Inc. VHF/UHF Transceiver Firmware
//
//	Copyright 2015 Simple Circuits Inc.
//
//	1200 Baud TNC
//
//	10/05/2016	Added CRC test on received packets
//
//----------------------------------------------------------------------------
#include "main.h"
#include <string.h>
#include <stdio.h>


//----------------------------------------------------------------------------
//
//	Defines
//
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//
//	Global variables
//
//----------------------------------------------------------------------------
uint16_t MixerPhase;					// mixer phase variables
uint16_t SpaceMixerPhase;				// mixer phase for space signal
int32_t  MarkIBitWindow[8];				// mark I sample buffer post mixer
int32_t	 MarkQBitWindow[8];				// mark Q
int32_t	 SpaceIBitWindow[8];			// space I sample buffer post mixer
int32_t	 SpaceQBitWindow[8];			// space Q
uint16_t WindowPtr;						// moving window integrator pointer
uint16_t Decimation1Cntr;				// decimation counter

uint32_t MarkMagnitude, SpaceMagnitude;
uint16_t CurrentBit, LastBit;			// current and last received data bit
uint8_t  UsartRxSR;						// USART receive shift register
uint16_t UsartRxBitCount;				// USART receive bit counter
uint16_t InSync;						// ax25 in-sync flag
uint8_t	 TncBuffer[300];
uint16_t TncBufferPtr;
uint16_t BitStuffCounter;				// counts ax25 consecutive 1's

volatile uint16_t GoodPackets, BadPackets;	// packet error counters

volatile uint16_t debugCRC1, debugCRC2, debugRx, debugLen;

//----------------------------------------------------------------------------
//
//	Function Prototypes
//
//----------------------------------------------------------------------------
void HdlcUsart(void);
uint32_t SquareAndSum(int32_t, int32_t);
void TNC(uint16_t);
uint16_t CrcCheck(uint8_t *, uint16_t);


//----------------------------------------------------------------------------
//
//	APRS routines
//
//----------------------------------------------------------------------------
void RunTnc1200(void)
{
	uint16_t i;
	int32_t  raw;
	uint32_t markI, markQ, spaceI, spaceQ;
	
	if ((RxInputPtr != RxOutputPtr))		// 9600 sps
	{
		raw = RxBuffer[RxOutputPtr++];
		RxOutputPtr &= sizeof(RxBuffer)/4 - 1;

		MarkIBitWindow[WindowPtr]  = Mixer(MixerPhase, raw);				// mark tone
		MarkQBitWindow[WindowPtr]  = Mixer(MixerPhase + 0x4000, raw);		// 90 degrees shift
		SpaceIBitWindow[WindowPtr] = Mixer(SpaceMixerPhase, raw);			// space tone
		SpaceQBitWindow[WindowPtr++] = Mixer(SpaceMixerPhase + 0x4000, raw);	// 90 degrees shift
		
		WindowPtr &= 7;
		MixerPhase += 8192;					// 65536 * 1200 / 9600 = 8192
		SpaceMixerPhase += 15019;			// 65536 * 2200 / 9600 = 15019
		
		//	Compute Fourier power coefficient for mark and space on previous 8 samples
		markI  = MarkIBitWindow[0];			// initialize with 1st samples
		markQ  = MarkQBitWindow[0];
		spaceI = SpaceIBitWindow[0];
		spaceQ = SpaceQBitWindow[0];
		for (i = 1; i < 8; i++)				// last 7 samples
		{
			markI  += MarkIBitWindow[i];
			markQ  += MarkQBitWindow[i];
			spaceI += SpaceIBitWindow[i];
			spaceQ += SpaceQBitWindow[i];
		}
		MarkMagnitude  += SquareAndSum(markI, markQ);
		MarkMagnitude  /= 2;
		SpaceMagnitude += SquareAndSum(spaceI, spaceQ);
		SpaceMagnitude /= 2;

		if (MarkMagnitude > (SpaceMagnitude + 0x1000))
		{
			if (CurrentBit)
			{
				CurrentBit = 0;
				Decimation1Cntr = 4;
			}
		}
		else if (SpaceMagnitude > (MarkMagnitude + 0x1000))
		{
			if (!CurrentBit)
			{
				CurrentBit = 1;
				Decimation1Cntr = 4;
			}
		}

		if (++Decimation1Cntr >= 8)			// integrate 8 samples, 1200 bps
		{
			Decimation1Cntr = 0;
			HdlcUsart();
		}
	}
}


//----------------------------------------------------------------------------
//
//	HDLC USART does flag detection and bit de-stuffing
//
//----------------------------------------------------------------------------
void HdlcUsart(void)
{
	uint16_t dBit;
	
	dBit = (CurrentBit != LastBit) ? 0 : 1;
	LastBit = CurrentBit;
	
	if (!InSync)				// not in sync
	{
		UsartRxSR >>= 1;		// LSB first
		if (dBit)
			UsartRxSR |= 0x80;

		if (UsartRxSR == 0x7E)	// flag byte
		{
			if (TncBufferPtr)
				TNC(TncBufferPtr);
			UsartRxBitCount = BitStuffCounter = TncBufferPtr = 0;
			InSync = TRUE;
		}
	}
	else if (BitStuffCounter == 5)	// in sync and 5 consecutive 1's
	{
		if (dBit)				// if dBit == 0, bit stuffing removal
		{
			InSync = FALSE;
			UsartRxSR >>= 1;
			UsartRxSR |= 0x80;
		}
	}
	else						// in sync
	{
		UsartRxSR >>= 1;		// LSB first
		if (dBit)
			UsartRxSR |= 0x80;

		if (++UsartRxBitCount == 8)
		{
			TncBuffer[TncBufferPtr++] = UsartRxSR;
			TncBufferPtr %= sizeof(TncBuffer);
			UsartRxBitCount = 0;
		}
	}
	
	BitStuffCounter = (dBit) ? BitStuffCounter + 1: 0;
}


//----------------------------------------------------------------------------
//
//	TNC processing
//
//----------------------------------------------------------------------------
#define	FEND			0xC0			// KISS TNC frame end
#define	FESC			0xDB			// KISS TNC frame escape
#define	TFEND			0xDC			// KISS TNC transposed frame end
#define	TFESC			0xDD			// KISS TNC transposed frame escape

void TNC(uint16_t len)
{
	uint16_t i;
	uint8_t chr;
	
	if ((len < 19) || (len > (255 + 19)))
		return;
	
	// check 16 bit FCS (frame check sequence) and toss bad packets
	if (CrcCheck(TncBuffer, len) == FALSE)
	{
		BadPackets++;
		return;
	}
	GoodPackets++;

	UART_SendByte(FEND);				// frame end
	UART_SendByte(0);					// command byte

	for (i = 0; i < len; i++)
	{
		chr = TncBuffer[i];
		if (chr == FEND)
		{
			UART_SendByte(FESC);		// send escape code
			UART_SendByte(TFEND);		// send translated FEND
		}
		else if (chr == FESC)
		{
			UART_SendByte(FESC);		// send escape code
			UART_SendByte(TFESC);		// send translated FESC
		}			
		else
			UART_SendByte(chr);
	}

	UART_SendByte(FEND);				// frame end
}


//----------------------------------------------------------------------------
//
//	Frame Check Sequence (FCS) computation using CCITT CRC16
//	Computes CRC, compares to received FCS and returns TRUE if good
//	FCS is per HDLC document ISO 3309
//	Polynomial: x**16 + x**12 + x**5 +1
//	FCS is transmitted bit 15 first, opposite of all other fields
//
//----------------------------------------------------------------------------
uint16_t CrcCheck(uint8_t * packet, uint16_t len)
{
	uint16_t i, j, crc;
	
	crc = 0xFFFF;		// CCITT seed
	for (i = len - 2; i; i--)
	{
		crc ^= (uint16_t) *packet++;
		for (j = 8; j; j--)
 		{
			if (crc & 0x001)
				crc = (crc >> 1) ^ 0x8408;	// CCITT polynomial
			else
				crc >>= 1;
		}
	}
	
	debugCRC1 = crc;
	crc ^= 0xFFFF;
	debugCRC2 = crc;
	
	i = (uint16_t) *packet++;
	i |= ((uint16_t) *packet) << 8;
	debugRx = i;
	debugLen = len;
	return (i == crc);
}


//
//	Verbose version for dumb terminals
//
//void TNC(uint16_t len)
//{
//	uint16_t i, aprsPtr;
//	uint8_t chr;
//	char str[512];
//	char *strPtr = (char *)str;
//	
//	if ((len < 19) || (len > (255 + 19)))
//		return;
//	
//	for (i = 7; i < 13; i++)			// from address
//		*strPtr++ = TncBuffer[i] >> 1;
//	*strPtr++ = '-';
//	*strPtr++ = ((TncBuffer[13] & 0x1E) >> 1) + '0';	// !! for now, only single digit
//	
//	*strPtr++ = '>';

//	for (i = 0; i < 6; i++)				// to address
//		*strPtr++ = TncBuffer[i] >> 1;
//	*strPtr++ = ((TncBuffer[6] & 0x1E) >> 1) + '0';	// !! for now, only single digit

//	UART_SendData(USART1, (uint8_t *)str, 15);
//	
//	chr = TncBuffer[13];				// get last byte of from address
//	aprsPtr = 14;
//	while (!(chr & 1) && (aprsPtr < 10*7) && (aprsPtr < len))
//	{
//		strPtr = (char *)str;
//		*strPtr++ = ',';
//		for (i = 0; i < 7; i++)
//		{
//			chr = TncBuffer[aprsPtr++];
//			*strPtr++ = chr >> 1;
//		}
//		UART_SendData(USART1,(uint8_t *)str, 8);
//	}
//	
//	if ((TncBuffer[aprsPtr] != 0x03) || (TncBuffer[aprsPtr + 1] != 0xF0))
//	{
//		UART_SendString(USART1, "\r\n");
//		return;
//	}

//	aprsPtr += 2;			// point past the CTL and PID bytes
//	UART_SendString(USART1, ": ");

//	if ((len - aprsPtr) < 258)
//	{
//		len -= aprsPtr;
//		while (len)
//		{
//			if (len > 63)
//			{
//				UART_SendData(USART1, &TncBuffer[aprsPtr], 63);
//				aprsPtr += 63;
//				len -= 63;
//			}
//			else
//			{
//				UART_SendData(USART1, &TncBuffer[aprsPtr], len);
//				len = 0;
//			}
//		}
//	}

//	UART_SendString(USART1, "\r\n");
//}


/* [] END OF FILE */
