//*****************************************************************************
//
//  Simple Circuits Inc. VHF/UHF Transceiver Firmware
//
//	Copyright 2015 Simple Circuits Inc.
//
//	DSP Functions
//
//	04/06/2015
//
//
//*****************************************************************************
#include "main.h"



//-----------------------------------------------------------------------------
//
//	Mixer routine, computes cosine and multiplies with input value.
//	Returns top 32 bits of signed result.
//	Enter with phase angle normalized to pi (unsigned Q15). 0 < x < 2
//	and the value to mix in Q31.
//	Sin computation using Taylor series approximation.
//	Total time including call and return is 0.980 usec
//
//-----------------------------------------------------------------------------
//
//	Taylor series constants to compute sin
//
const uint32_t K7 = 0xD9A5A663;
const uint32_t K5 = 0x28CD78CC;		// k5 = (2^28-1)*pi^5/120; Q28
const uint32_t K3 = 0xEB544320;		// k3 = -(2^26-1)*pi^3/6; Q26
const uint32_t K1 = 0x03243F6A;		// k1 = (2^24-1)* pi;  Q24

__asm int32_t Mixer(uint16_t x, int32_t sig)
{
		// r0 = x, r1 = input signal
		PUSH	{r1-r5}				// save the registers that are used
		TST		r0, #0x8000			// r0 = x, unsigned Q15
		BEQ		LT1					// brif less than 1

		SUB		r0, r0, #0xC000		// x = x - 1.5pi
		B		Mix1

LT1		RSB		r0, r0, #0x4000		// x < .5pi, x = .5pi - j

Mix1	MUL		r4, r0, r0			// r4 = xsq = x*x; Q30
		LDR		r5, =__cpp(K7)		// k7 = -(2^30-1)*pi^7/5040; Q30
		SMULL	r2, r3, r4, r5		// r3 = sine = (xsq * k7) >> 32;  Q28
		LDR		r5, =__cpp(K5)		// k5 = (2^28-1)*pi^5/120; Q28
		ADD		r3, r3, r5			// r3 = sine += k5;  Q28
		SMULL	r2, r3, r4, r3		// r3 = sine = (sine * xsq) >> 32; Q26
		LDR		r5, =__cpp(K3)		// k3 = -(2^26-1)*pi^3/6; Q26
		ADD		r3, r3, r5			// r3 = sine += k3; Q26
		SMULL	r2, r3, r4, r3		// r3 = sine = (sine * xsq) >> 32; Q24
		LDR		r5, =__cpp(K1)		// k1 = (2^24-1)* pi;  Q24
		ADD		r3, r3, r5			// r3 = sine += k1; Q24
		MOV		r4, r0, LSL #16		// r0 = x in Q31 format	
		SMULL	r2, r3, r4, r3		// r3 = sine = (sine * x) >> 32;  Q23

		MOV		r3, r3, LSL #7		// make Q30 format

		SMULL	r1, r0, r3, r1		// r0 = (sine * sig) >> 32; Q29
		POP		{r1-r5}				// restore the registers
		BX		lr
}


//-----------------------------------------------------------------------------
//
//	uint32_t SquareAndSum(int32, int32)
//
//	Squares inputs and then sums
//	Performs integer (x*x + y*y) in assembly
//	Inputs: R0 = int32, R1 = int32 
//	Output: R0 = uint32 integer return
//	Does not overflow for any signed 16 bit input values
//
//-----------------------------------------------------------------------------
__asm uint32_t SquareAndSum(int32_t x, int32_t y)
{
	push	{r1}
	mul		r0, r0, r0			/* r0 = r0*r0			*/
	mla		r0, r1, r1, r0		/* r0 = r0*r0 + r1*r1	*/

	pop		{r1}
	bx		lr
}


//
//	FIR BPF design by Matlab
//	Equiripple, 63rd order, Fs 9600, Fstop1 200, Fpass1 525, Fpass2 2200, Fstop2 2500
//
#define NORM			0xCFFFFFFF		// normalize coeffients to 1
const int32_t FIR_BPF_COEFF[64] = {
NORM *  0.001022519520192, NORM * -0.001516432005487, NORM * -0.006455335352981, NORM * -0.006332672844795,
NORM *  0.002876619759216, NORM *  0.009875957075582, NORM *  0.001406514850263, NORM * -0.013875836167260,
NORM * -0.013880097570699, NORM *  0.001703213020836, NORM *  0.008873944733995, NORM *  0.000414626910810, 
NORM * -0.000049889075798, NORM *  0.017994556173206, NORM *  0.027474231346852, NORM *  0.011120387316438, 
NORM * -0.001712896898751, NORM *  0.014273518250060, NORM *  0.027745512443189, NORM *  0.003230002792828, 
NORM * -0.026543180732169, NORM * -0.012600526931828, NORM *  0.012544874860193, NORM * -0.017968059172416,
NORM * -0.074138105071940, NORM * -0.062866119035709, NORM * -0.004411130689271, NORM * -0.030773304690003,
NORM * -0.147676523722668, NORM * -0.152306099536506, NORM *  0.080852566235942, NORM *  0.351108478786903, 
NORM *  0.351108478786903, NORM *  0.080852566235942, NORM * -0.152306099536506, NORM * -0.147676523722668,
NORM * -0.030773304690003, NORM * -0.004411130689271, NORM * -0.062866119035709, NORM * -0.074138105071940,
NORM * -0.017968059172416, NORM *  0.012544874860193, NORM * -0.012600526931828, NORM * -0.026543180732169,
NORM *  0.003230002792828, NORM *  0.027745512443189, NORM *  0.014273518250060, NORM * -0.001712896898751,
NORM *  0.011120387316438, NORM *  0.027474231346852, NORM *  0.017994556173206, NORM * -0.000049889075798,
NORM *  0.000414626910810, NORM *  0.008873944733995, NORM *  0.001703213020836, NORM * -0.013880097570699,
NORM * -0.013875836167260, NORM *  0.001406514850263, NORM *  0.009875957075582, NORM *  0.002876619759216, 
NORM * -0.006332672844795, NORM * -0.006455335352981, NORM * -0.001516432005487, NORM *  0.001022519520192};


//
//	FIR 32bit
//	Enter with pointer to base of data buffer, index into buffer, point to coefficients, length in words
//	Return FIR output as s32
//
__asm int32_t Fir(int32_t *signal, uint16_t rxptr, const int32_t *coeff, uint16_t len)
{
		// r0 = pointer to sig buffer, r1 = pointer into buffer
		// r2 = pointer to coefficients, r3 = FIR length
		push	{r1-r8}					// save the registers that are used
		mov		r8, r3					// save FIR length
		mov		r3, r3, lsl #2			// compute length in bytes, not words
		mov		r4, r1, lsl #2			// signal pointer multiplied by 4 to convert bytes to words
		mov		r5, r0					// data buffer base address
		mov		r0, #0					// clear the accumulator
		mov		r1, #0

FIR2LP	ldr		r6, [r5, r4]			// get data word
		ldr		r7, [r2], #4			// get FIR coefficient
		smlal	r1, r0, r6, r7			// multiply and accumulate
		add		r4, r4, #4				// increment data pointer by 4 bytes (to next word)
		cmp		r4, r3
		bne		FIRLP2					// branch if no wrap around needed
		mov		r4, #0
		
FIRLP2	subs	r8, #1
		bne		FIR2LP
		
		// return with r0 = FIR output
		pop		{r1-r8}					// restore the registers
		bx		lr
}


