//------------------------------------------------------------------------------
//
//  Simple Circuits Inc. VHF/UHF Transceiver Firmware
//
//	Copyright 2015 Simple Circuits Inc.
//
//	Receiver Processing Functions
//
//	04/13/2015
//
//------------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include "main.h"
#include "radio.h"


//----------------------------------------------------------------------------
//
//	Global Variables
//
//----------------------------------------------------------------------------
int32_t	RxBuffer[64];						// receive samples for demodulators
uint16_t RxInputPtr;						// input pointer into rx buffer
uint16_t RxOutputPtr;						// output pointer into rx buffer
uint8_t RX_SpiStatus;
uint16_t RX_SpiError;
extern uint8_t const RF_SETTINGS[47];
extern uint8_t const EXTENDED_RF_SETTINGS[56];

uint16_t Decimator = 4;
int32_t Intg1,Intg2,Intg3,Intg4;			// integrators
int32_t Comb1,Comb2,Comb3,Comb4;			// comb filter summed outputs
int32_t Comb1Delay,Comb2Delay,Comb3Delay,Comb4Delay;	// comb filter delays


//----------------------------------------------------------------------------
//
//	Local Function Prototypes
//
//----------------------------------------------------------------------------
void  RX_SpiReady(void);
void  RX_WriteRegister(uint16_t, uint8_t);
uint8_t RX_ReadRegister(uint16_t);
uint8_t RX_CommandStrobe(uint8_t reg);
uint8_t RX_WriteSpi(uint8_t);


//----------------------------------------------------------------------------
//
//	Receive data ready interrupt routine
//	38.4 KHz, 26 usec
//	Executes in 0.5 usec, 168 MHz CPU
//
//----------------------------------------------------------------------------
void RadioISR(void)
{
	int32_t raw;
	
	// 4th order CIC, decimate by 4, differential delay = 1
	// bit growth is N*log2(4*1) = 4*log2(4) = 8 bits
	Intg1 += (int8_t) SPI2->DR;			// read sample from last SPI transfer
	SPI2->DR = 0;						// get next sample going
	Intg2 += Intg1;
	Intg3 += Intg2;
	Intg4 += Intg3;

	if (--Decimator == 0)				// comb filter section
	{
		Decimator = 4;
		
		Comb1 = Intg4 - Comb1Delay;
		Comb2 = Comb1 - Comb2Delay;
		Comb3 = Comb2 - Comb3Delay;
		Comb4 = Comb3 - Comb4Delay;
		Comb1Delay = Intg4;
		Comb2Delay = Comb1;
		Comb3Delay = Comb2;
		Comb4Delay = Comb3;
		RxBuffer[RxInputPtr++] = Comb4;
		RxInputPtr &= sizeof(RxBuffer)/4 - 1;

		TP5_LO();
		// CIC output to BPF input, removes CTCSS tones and audio above 2.2 KHz
		FirStage[FirStagePtr] = Comb4;		
		raw = Fir(FirStage, FirStagePtr++, FIR_BPF_COEFF, 64);
		FirStagePtr &= 63;
		TP5_HI();
		
		if (!SW1())
			raw = Comb4 >> ((7-Volume) + 4);
		else
			raw >>= ((7-Volume) + 3);
		DAC->DHR12R1 = 0x0800 ^ raw;	// load the DAC
	}
}


//----------------------------------------------------------------------------
//
//	Compute frequency settings and load radio registers
//	Frequency is in Hertz
//	frequency (144MHz) = FREQ * 32 MHz / (65536 * LO_DIVIDER)
//
//----------------------------------------------------------------------------
void SetFrequency(uint64_t frq)
{
	uint8_t rh,rm,rl,rb;
	uint64_t loDiv;
	
	RADIO_NSS1();		// prepare to write radio registers, SS was low
	
	if (frq <= 160000000)
	{
		loDiv = 24;
		rb = 0x0B;								// selects prescaler of 24
	}
	else if (frq <= 192000000)
	{
		loDiv = 20;
		rb = 0x0A;								// selects prescaler of 20
	}
	else if (frq <= 240000000)
	{
		loDiv = 16;
		rb = 0x08;								// selects prescaler of 16
	}
	else if (frq <= 365000000)
	{
		loDiv = 12;
		rb = 0x06;								// selects prescaler of 12
	}
	else if (frq <= 648000000)
	{
		loDiv = 8;
		rb = 0x04;								// selects prescaler of 8
	}
	else if (frq <= 999999999)
	{
		loDiv = 4;
		rb = 0x02;								// selects prescaler of 4
	}
	else
	{
		frq = 144390000;
		loDiv = 24;
		rb = 0x0B;								// selects prescaler of 24
	}

	frq = (frq * 65536 * loDiv) / 32000000;		// compute FREQ register value
	rh = (uint8_t) (frq >> 16);
    rm = (uint8_t) (frq >> 8) ;
    rl = (uint8_t)  frq;

	__disable_irq();
	RX_CommandStrobe(RADIO_CMD_SIDLE);			// stop RX mode
	RX_WriteRegister(CC112X_FREQ2, rh);
	RX_WriteRegister(CC112X_FREQ1, rm);
	RX_WriteRegister(CC112X_FREQ0, rl);
	RX_WriteRegister(CC112X_FS_CFG, rb);
	RX_CommandStrobe(RADIO_CMD_SRX);			// enable RX mode
	__enable_irq();								// enable all interrupts again
	
	RX_CommandStrobe(RADIO_CMD_SRX);			// enable RX mode again
	RX_SpiReady();
	RX_WriteSpi(RADIO_CMD_RDEXTENED_REG | 0x40);	// extended register access command, burst enabled
	RX_WriteSpi((uint8_t) CC112X_CFM_RX_DATA_OUT);	// register write address
}


//----------------------------------------------------------------------------
//
//	Send byte to RX radio
//
//----------------------------------------------------------------------------
uint8_t RX_WriteSpi(uint8_t data_in)
{
	SPI2->DR = data_in;
	while ( !(SPI2->SR & SPI_FLAG_TXE) ) ;
	while (  (SPI2->SR & SPI_FLAG_BSY) ) ;
	while ( !(SPI2->SR & SPI_FLAG_RXNE) ) ;

	return SPI2->DR;
}


//----------------------------------------------------------------------------
//
//	Write RX register with single data byte
//
//----------------------------------------------------------------------------
void RX_WriteRegister(uint16_t reg, uint8_t data)
{
	RX_SpiReady();
	if (reg > 0x00FF)
		RX_WriteSpi(RADIO_CMD_WREXTENED_REG);	// extended register access command, no burst
	RX_WriteSpi((uint8_t) reg);					// write register address
	RX_WriteSpi(data);
	RADIO_NSS1();
}


//----------------------------------------------------------------------------
//
//	Read RX register single data byte
//
//----------------------------------------------------------------------------
uint8_t RX_ReadRegister(uint16_t reg)
{
	uint8_t returnData;
	
	RX_SpiReady();
	if (reg > 0x00FF)
	{
		RX_SpiStatus = RX_WriteSpi(RADIO_CMD_RDEXTENED_REG);	// extended register access command, no burst
		RX_WriteSpi((uint8_t) reg);								// register write address
	}
	else
		RX_SpiStatus = RX_WriteSpi(0x80 | (uint8_t) reg);		// register write address
	returnData = RX_WriteSpi(0);
	RADIO_NSS1();
	
	return returnData;
}


//----------------------------------------------------------------------------
//
//	Write RX command strobe register
//
//----------------------------------------------------------------------------
uint8_t RX_CommandStrobe(uint8_t reg)
{
	uint8_t returnData;
	
	RX_SpiReady();
	returnData = RX_WriteSpi(reg);		// write command strobe register
	RADIO_NSS1();
	return (returnData);	
}


//----------------------------------------------------------------------------
//
//	Wait for RX SPI ready, if time out and restart CS
//
//----------------------------------------------------------------------------
void RX_SpiReady(void)
{
	uint16_t i = 5000;					// 200 was good, 100 was too short
	
	RADIO_NSS0();
	while(HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_14))
	{
		if (--i == 0)
		{
	        RADIO_NSS1();		// time out, remove CS, delay and restart CS
			Delay1Ms(2);
        	RADIO_NSS0();
			i = 5000;
			RX_SpiError++;
		}
	}
}


//----------------------------------------------------------------------------
//
//	Manual calibration of the CC1120 VCO per TI Errata sheet SWRZ039B
//
//----------------------------------------------------------------------------
#define VCDAC_START_OFFSET	2
#define FS_VCO2_INDEX		0
#define FS_VCO4_INDEX		1
#define FS_CHP_INDEX		2

void ManualCalibration(void)
{
	uint8_t rx1_original_fs_cal2;
	uint8_t rx1_calResults_for_vcdac_start_high[3];
	uint8_t rx1_calResults_for_vcdac_start_mid[3];

	// 1) Set VCO cap-array to 0 (FS_VCO2 = 0x00)
	RX_WriteRegister(CC112X_FS_VCO2, 0x00);
	
	// 2) Start with high VCDAC (original VCDAC_START + 2):
	rx1_original_fs_cal2 = RX_ReadRegister(CC112X_FS_CAL2);
	RX_WriteRegister(CC112X_FS_CAL2, rx1_original_fs_cal2 + VCDAC_START_OFFSET);


	// 3) Calibrate and wait for calibration to be done (radio back in IDLE state)
	RX_CommandStrobe(RADIO_CMD_SCAL);
	while (RX_ReadRegister(CC112X_MARCSTATE) != 0x41) ;

	// 4) Read FS_VCO2, FS_VCO4 and FS_CHP register obtained with high VCDAC_START value
	rx1_calResults_for_vcdac_start_high[FS_VCO2_INDEX] = RX_ReadRegister(CC112X_FS_VCO2);
	rx1_calResults_for_vcdac_start_high[FS_VCO4_INDEX] = RX_ReadRegister(CC112X_FS_VCO4);
	rx1_calResults_for_vcdac_start_high[FS_CHP_INDEX]  = RX_ReadRegister(CC112X_FS_CHP);

	// 5) Set VCO cap-array to 0 (FS_VCO2 = 0x00)
	RX_WriteRegister(CC112X_FS_VCO2, 0x00);

	// 6) Continue with mid VCDAC (original VCDAC_START):
	RX_WriteRegister(CC112X_FS_CAL2, rx1_original_fs_cal2);

	// 7) Calibrate and wait for calibration to be done (radio back in IDLE state)
	RX_CommandStrobe(RADIO_CMD_SCAL);
	while (RX_ReadRegister(CC112X_MARCSTATE) != 0x41) ;

	// 8) Read FS_VCO2, FS_VCO4 and FS_CHP register obtained with mid VCDAC_START value
	rx1_calResults_for_vcdac_start_mid[FS_VCO2_INDEX] = RX_ReadRegister(CC112X_FS_VCO2);
	rx1_calResults_for_vcdac_start_mid[FS_VCO4_INDEX] = RX_ReadRegister(CC112X_FS_VCO4);
	rx1_calResults_for_vcdac_start_mid[FS_CHP_INDEX]  = RX_ReadRegister(CC112X_FS_CHP);

	// 9) Write back highest FS_VCO2 and corresponding FS_VCO and FS_CHP result
	if (rx1_calResults_for_vcdac_start_high[FS_VCO2_INDEX] > rx1_calResults_for_vcdac_start_mid[FS_VCO2_INDEX])
	{
		RX_WriteRegister(CC112X_FS_VCO2, rx1_calResults_for_vcdac_start_high[FS_VCO2_INDEX]);
		RX_WriteRegister(CC112X_FS_VCO4, rx1_calResults_for_vcdac_start_high[FS_VCO4_INDEX]);
		RX_WriteRegister(CC112X_FS_CHP, rx1_calResults_for_vcdac_start_high[FS_CHP_INDEX]);
	}
	else
	{
		RX_WriteRegister(CC112X_FS_VCO2, rx1_calResults_for_vcdac_start_mid[FS_VCO2_INDEX]);
		RX_WriteRegister(CC112X_FS_VCO4, rx1_calResults_for_vcdac_start_mid[FS_VCO4_INDEX]);
		RX_WriteRegister(CC112X_FS_CHP, rx1_calResults_for_vcdac_start_mid[FS_CHP_INDEX]);
	}
}


//----------------------------------------------------------------------------
//
//	Calibration of the CC1120 RC Oscillator
//
//----------------------------------------------------------------------------
void CalibrateRCOscillator(void)
{
    uint8_t temp;

    // Read current register value
	temp = RX_ReadRegister(CC112X_WOR_CFG0);
	
    // Mask register bit fields and write new values
    temp = (temp & 0xF9) | (0x02 << 1);

    // Write new register value
	RX_WriteRegister(CC112X_WOR_CFG0, temp);
	
    // Strobe IDLE to calibrate the RCOSC
    RX_CommandStrobe(RADIO_CMD_SIDLE);

    // Disable RC calibration
    temp = (temp & 0xF9) | (0x00 << 1);
	RX_WriteRegister(CC112X_WOR_CFG0, temp);
}


//----------------------------------------------------------------------------
//
//	Initialize CC1120 radio configuration registers
//
//----------------------------------------------------------------------------
void ConfigureRadio(void)
{
	uint16_t i;

	RADIO_NSS1();
	
	Delay1Ms(500);						// power up settling
	RADIO_RST0();						// reset the chip
	Delay1Ms(100);
	RADIO_RST1();;
	Delay1Ms(100);
	
	// Initialize radio
	RX_SpiReady();
	RX_WriteSpi(0x40);
	for (i = 0; i < sizeof(RF_SETTINGS); i++)
		RX_WriteSpi(RF_SETTINGS[i]);
	RADIO_NSS1();
	
	Delay1Ms(10);

	RX_SpiReady();
	RX_WriteSpi(0x6F);					// extended register access command, burst mode
	RX_WriteSpi(0x00);					// extended register address from 0
	for (i = 0; i < sizeof(EXTENDED_RF_SETTINGS); i++)
		RX_WriteSpi(EXTENDED_RF_SETTINGS[i]);
	RADIO_NSS1();

	Delay1Ms(250);

	ManualCalibration();				// calibrate FS VCO
	Delay1Ms(100);
	
	SetFrequency(DEFAULT_START_FRQ);
	
	RX_CommandStrobe(RADIO_CMD_SRX);	// enable RX mode
	
	RX_WriteRegister(CC112X_EXT_CTRL, 0);			// disable burst address increment

	RX_SpiReady();
	RX_WriteSpi(RADIO_CMD_RDEXTENED_REG | 0x40);	// extended register access command, burst enabled
	RX_WriteSpi((uint8_t) CC112X_CFM_RX_DATA_OUT);	// register write address
}


